"""A package to specifically handle crystal structures and microstructures.

   The crystal package has three main modules:

   * :py:mod:`~pymicro.crystal.lattice`
   * :py:mod:`~pymicro.crystal.microstructure`
   * :py:mod:`~pymicro.crystal.texture`

   .. moduleauthor:: Henry Proudhon <henry.proudhon@gmail.com>

"""
