"""A package to play with X-rays

   The xray package currently has five modules:

   * dct (experimental)
   * detectors
   * fitting
   * laue
   * xray_utils

   .. warning::

        The documentation of this package is still very poor, use with care...
        and add some love to the docs if you have time!

   .. moduleauthor:: Henry Proudhon <henry.proudhon@mines-paristech.fr>

"""
